/*
Module : PropVariantWrapper.h
Purpose: Defines the interface for a set of C++ wrapper classes of the PROPVARIANT SDK structure. The 
         classes are called CComPropVariant and CComPropVariant2 and are modelled upon the ATL 
         CComVariant class.
History: PJN / 05-05-2014 1. Initial public release
         PJN / 22-09-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean
                          that the code will now only compile on VC 2017 or later.
         PJN / 20-04-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 31-10-2021 1. Updated copyright details.
         PJN / 19-02-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations

Copyright (c) 2014 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


////////////////////////// Macros / Defines ///////////////////////////////////

#pragma once

#ifndef __PROPVARIANTWRAPPER_H__
#define __PROPVARIANTWRAPPER_H__

#pragma comment(lib, "propsys.lib") //Automatically link in the propsys dll



////////////////////////// Includes ///////////////////////////////////////////

#include <propvarutil.h>


////////////////////////// Classes ////////////////////////////////////////////

namespace ATL
{


//A wrapper class for PROPVARIANT with automatic lifetime management of the resource
//via PropVariantClear in its destructor
class CComPropVariant : public PROPVARIANT
{
public:
//Constructors / Destructors
  CComPropVariant() noexcept
  {
    //Make sure that propvariant data is initialized to 0
    memset(this, 0, sizeof(PROPVARIANT));
    PropVariantInit(this);
  }

  ~CComPropVariant() noexcept
  {
    const HRESULT hr{Clear()};
#pragma warning(suppress: 26477)
    ATLASSERT(SUCCEEDED(hr));
    UNREFERENCED_PARAMETER(hr);
  }

  CComPropVariant(_In_ const PROPVARIANT& varSrc) ATLVARIANT_THROW()
  {
    vt = VT_EMPTY;
    InternalCopy(&varSrc);
  }

  CComPropVariant(_In_ const CComPropVariant& varSrc) ATLVARIANT_THROW()
  {
    vt = VT_EMPTY;
    InternalCopy(&varSrc);
  }

#pragma warning(suppress: 26439 26495)
  CComPropVariant(_In_ CComPropVariant&& varSrc)
  {
    Checked::memcpy_s(this, sizeof(CComPropVariant), &varSrc, sizeof(CComPropVariant));
    memset(&varSrc, 0, sizeof(PROPVARIANT));
    PropVariantInit(&varSrc);
  }

  CComPropVariant(_In_z_ LPCOLESTR lpszSrc) ATLVARIANT_THROW()
  {
    vt = VT_EMPTY;
    *this = lpszSrc;
  }

  CComPropVariant(_In_z_ LPCSTR lpszSrc) ATLVARIANT_THROW()
  {
    vt = VT_EMPTY;
    *this = lpszSrc;
  }

  CComPropVariant(_In_ const FILETIME& ft) noexcept
  {
    vt = VT_FILETIME;
    filetime = ft;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ bool bSrc) noexcept
  {
    vt = VT_BOOL;
    boolVal = bSrc ? VARIANT_TRUE : VARIANT_FALSE;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ int nSrc, _In_ VARTYPE vtSrc = VT_I4) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_I4) || (vtSrc == VT_INT));
    if ((vtSrc == VT_I4) || (vtSrc == VT_INT))
    {
      vt = vtSrc;
      intVal = nSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ BYTE nSrc) noexcept
  {
    vt = VT_UI1;
    bVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ short nSrc) noexcept
  {
    vt = VT_I2;
    iVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ long nSrc, _In_ VARTYPE vtSrc = VT_I4) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_I4) || (vtSrc == VT_ERROR));
    if (vtSrc == VT_I4)
    {
      vt = vtSrc;
      lVal = nSrc;
    }
    else if (vtSrc == VT_ERROR)
    {
      vt = vtSrc;
      scode = nSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ float fltSrc) noexcept
  {
    vt = VT_R4;
    fltVal = fltSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ double dblSrc, _In_ VARTYPE vtSrc = VT_R8) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_R8) || (vtSrc == VT_DATE));
    if (vtSrc == VT_R8)
    {
      vt = vtSrc;
      dblVal = dblSrc;
    }
    else if (vtSrc == VT_DATE)
    {
      vt = vtSrc;
      date = dblSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ LARGE_INTEGER nSrc) noexcept
  {
    vt = VT_I8;
    hVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ ULARGE_INTEGER nSrc) noexcept
  {
    vt = VT_UI8;
    uhVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ CY cySrc) noexcept
  {
    vt = VT_CY;
    cyVal.Hi = cySrc.Hi;
    cyVal.Lo = cySrc.Lo;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ IDispatch* pSrc)
  {
    vt = VT_DISPATCH;
    pdispVal = pSrc;

    //Need to AddRef as PropVariantClear will Release
    if (pdispVal != nullptr)
      pdispVal->AddRef();
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ IUnknown* pSrc)
  {
    vt = VT_UNKNOWN;
    punkVal = pSrc;

    //Need to AddRef as PropVariantClear will Release
    if (punkVal != nullptr)
      punkVal->AddRef();
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ IStream* pSrc, _In_ VARTYPE vtSrc = VT_STREAM) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_STREAM) || (vtSrc == VT_STREAMED_OBJECT));
    if ((vtSrc == VT_STREAM) || (vtSrc == VT_STREAMED_OBJECT))
    {
      vt = vtSrc;
      pStream = pSrc;

      //Need to AddRef as PropVariantClear will Release
      if (pStream != nullptr)
        pStream->AddRef();
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ IStorage* pSrc, _In_ VARTYPE vtSrc = VT_STORAGE) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_STORAGE) || (vtSrc == VT_STORED_OBJECT));
    if ((vtSrc == VT_STORAGE) || (vtSrc == VT_STORED_OBJECT))
    {
      vt = vtSrc;
      pStorage = pSrc;

      //Need to AddRef as PropVariantClear will Release
      if (pStorage != nullptr)
        pStorage->AddRef();
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ char cSrc) noexcept
  {
    vt = VT_I1;
    cVal = cSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ unsigned short nSrc) noexcept
  {
    vt = VT_UI2;
    uiVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ unsigned long nSrc) noexcept
  {
    vt = VT_UI4;
    ulVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ unsigned int nSrc, _In_ VARTYPE vtSrc = VT_UI4) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_UI4) || (vtSrc == VT_UINT));
    if ((vtSrc == VT_UI4) || (vtSrc == VT_UINT))
    {
      vt = vtSrc;
      uintVal = nSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ const CComBSTR& bstrSrc) ATLVARIANT_THROW()
  {
    vt = VT_EMPTY;
    *this = bstrSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_ const SAFEARRAY* pSrc) ATLVARIANT_THROW()
  {
    vt = VT_EMPTY;
    *this = pSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ BYTE* pbSrc) noexcept
  {
    vt = VT_UI1|VT_BYREF;
    pbVal = pbSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ short* pnSrc) noexcept
  {
    vt = VT_I2|VT_BYREF;
    piVal = pnSrc;
  }

#ifdef _NATIVE_WCHAR_T_DEFINED
#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ USHORT* pnSrc) noexcept
  {
    vt = VT_UI2|VT_BYREF;
    puiVal = pnSrc;
  }
#endif //#ifdef _NATIVE_WCHAR_T_DEFINED

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ int* pnSrc) noexcept
  {
    vt = VT_I4|VT_BYREF;
    pintVal = pnSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ UINT* pnSrc) noexcept
  {
    vt = VT_UI4|VT_BYREF;
    puintVal = pnSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ long* pnSrc) noexcept
  {
    vt = VT_I4|VT_BYREF;
    plVal = pnSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ ULONG* pnSrc) noexcept
  {
    vt = VT_UI4|VT_BYREF;
    pulVal = pnSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ CY* pcySrc) noexcept
  {
    vt = VT_CY|VT_BYREF;
    pcyVal = pcySrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ float* pfSrc) noexcept
  {
    vt = VT_R4|VT_BYREF;
    pfltVal = pfSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant(_In_opt_ double* pdblSrc, _In_ VARTYPE vtSrc = VT_R8|VT_BYREF) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == (VT_R8|VT_BYREF)) || (vtSrc == (VT_DATE|VT_BYREF)));
    if (vtSrc == (VT_R8|VT_BYREF))
    {
      vt = vtSrc;
      pdblVal = pdblSrc;
    }
    else if (vtSrc == (VT_DATE|VT_BYREF))
    {
      vt = vtSrc;
      pdate = pdblSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26456)
  CComPropVariant& operator=(_In_ const CComPropVariant& varSrc) ATLVARIANT_THROW()
  {
    if (this != &varSrc)
      InternalCopy(&varSrc);
    return *this;
  }

#pragma warning(suppress: 26439 26456)
  CComPropVariant& operator=(_In_ CComPropVariant&& varSrc)
  {
    Checked::memcpy_s(this, sizeof(CComPropVariant), &varSrc, sizeof(CComPropVariant));
    memset(&varSrc, 0, sizeof(PROPVARIANT));
    PropVariantInit(&varSrc);
    return *this;
  }

#pragma warning(suppress: 26434 26487)
  CComPropVariant& operator=(_In_ const PROPVARIANT& varSrc) ATLVARIANT_THROW()
  {
    if (this != &varSrc)
      InternalCopy(&varSrc);
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ const CComBSTR& bstrSrc) ATLVARIANT_THROW()
  {
    ClearThrow();

    vt = VT_BSTR;
    bstrVal = bstrSrc.Copy();

    if ((bstrVal == nullptr) && (bstrSrc.m_str != nullptr))
    {
      vt = VT_ERROR;
      scode = E_OUTOFMEMORY;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_OUTOFMEMORY);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_z_ LPCOLESTR lpszSrc) ATLVARIANT_THROW()
  {
    ClearThrow();

    vt = VT_BSTR;
    bstrVal = SysAllocString(lpszSrc);

    if ((bstrVal == nullptr) && (lpszSrc != nullptr))
    {
      vt = VT_ERROR;
      scode = E_OUTOFMEMORY;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_OUTOFMEMORY);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_z_ LPCSTR lpszSrc) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26494 26496 26477)
    USES_CONVERSION_EX;
    ClearThrow();

    vt = VT_BSTR;
#pragma warning(suppress: 26472 26477)
    bstrVal = SysAllocString(A2COLE_EX(lpszSrc, _ATL_SAFE_ALLOCA_DEF_THRESHOLD));
    if ((bstrVal == nullptr) && (lpszSrc != nullptr))
    {
      vt = VT_ERROR;
      scode = E_OUTOFMEMORY;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_OUTOFMEMORY);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
#pragma warning(suppress: 26487 26489)
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ bool bSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_BOOL)
    {
      ClearThrow();
      vt = VT_BOOL;
    }
    boolVal = bSrc ? VARIANT_TRUE : VARIANT_FALSE;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ const FILETIME& nSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_FILETIME)
    {
      ClearThrow();
      vt = VT_FILETIME;
    }
    filetime = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ int nSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_I4)
    {
      ClearThrow();
      vt = VT_I4;
    }
    intVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ BYTE nSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_UI1)
    {
      ClearThrow();
      vt = VT_UI1;
    }
    bVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ short nSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_I2)
    {
      ClearThrow();
      vt = VT_I2;
    }
    iVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ long nSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_I4)
    {
      ClearThrow();
      vt = VT_I4;
    }
    lVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ float fltSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_R4)
    {
      ClearThrow();
      vt = VT_R4;
    }
    fltVal = fltSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ double dblSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_R8)
    {
      ClearThrow();
      vt = VT_R8;
    }
    dblVal = dblSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ CY cySrc) ATLVARIANT_THROW()
  {
    if (vt != VT_CY)
    {
      ClearThrow();
      vt = VT_CY;
    }
    cyVal.Hi = cySrc.Hi;
    cyVal.Lo = cySrc.Lo;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ char cSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_I1)
    {
      ClearThrow();
      vt = VT_I1;
    }
    cVal = cSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ unsigned short nSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_UI2)
    {
      ClearThrow();
      vt = VT_UI2;
    }
    uiVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ unsigned long nSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_UI4)
    {
      ClearThrow();
      vt = VT_UI4;
    }
    ulVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ unsigned int nSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_UI4)
    {
      ClearThrow();
      vt = VT_UI4;
    }
    uintVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ LARGE_INTEGER nSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_I8)
    {
      ClearThrow();
      vt = VT_I8;
    }
    hVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ ULARGE_INTEGER nSrc) ATLVARIANT_THROW()
  {
    if (vt != VT_UI8)
    {
      ClearThrow();
      vt = VT_UI8;
    }
    uhVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_Inout_opt_ IDispatch* pSrc) ATLVARIANT_THROW()
  {
    ClearThrow();

    vt = VT_DISPATCH;
    pdispVal = pSrc;

    //Need to AddRef as PropVariantClear will Release
    if (pdispVal != nullptr)
      pdispVal->AddRef();

    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_Inout_opt_ IUnknown* pSrc) ATLVARIANT_THROW()
  {
    ClearThrow();

    vt = VT_UNKNOWN;
    punkVal = pSrc;

    //Need to AddRef as PropVariantClear will Release
    if (punkVal != nullptr)
      punkVal->AddRef();

    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ BYTE* pbSrc) ATLVARIANT_THROW()
  {
    if (vt != (VT_UI1|VT_BYREF))
    {
      ClearThrow();
      vt = VT_UI1|VT_BYREF;
    }
    pbVal = pbSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ short* pnSrc) ATLVARIANT_THROW()
  {
    if (vt != (VT_I2|VT_BYREF))
    {
      ClearThrow();
      vt = VT_I2|VT_BYREF;
    }
    piVal = pnSrc;
    return *this;
  }

#ifdef _NATIVE_WCHAR_T_DEFINED
#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ USHORT* pnSrc) ATLVARIANT_THROW()
  {
    if (vt != (VT_UI2|VT_BYREF))
    {
      ClearThrow();
      vt = VT_UI2|VT_BYREF;
    }
    puiVal = pnSrc;
    return *this;
  }
#endif //#ifdef _NATIVE_WCHAR_T_DEFINED

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ int* pnSrc) ATLVARIANT_THROW()
  {
    if (vt != (VT_I4|VT_BYREF))
    {
      ClearThrow();
      vt = VT_I4|VT_BYREF;
    }
    pintVal = pnSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ UINT* pnSrc) ATLVARIANT_THROW()
  {
    if (vt != (VT_UI4|VT_BYREF))
    {
      ClearThrow();
      vt = VT_UI4|VT_BYREF;
    }
    puintVal = pnSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ long* pnSrc) ATLVARIANT_THROW()
  {
    if (vt != (VT_I4|VT_BYREF))
    {
      ClearThrow();
      vt = VT_I4|VT_BYREF;
    }
    plVal = pnSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ ULONG* pnSrc) ATLVARIANT_THROW()
  {
    if (vt != (VT_UI4|VT_BYREF))
    {
      ClearThrow();
      vt = VT_UI4|VT_BYREF;
    }
    pulVal = pnSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ CY* pVal) ATLVARIANT_THROW()
  {
    if (vt != (VT_CY|VT_BYREF))
    {
      ClearThrow();
      vt = VT_CY|VT_BYREF;
    }
    pcyVal = pVal;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ float* pfSrc) ATLVARIANT_THROW()
  {
    if (vt != (VT_R4|VT_BYREF))
    {
      ClearThrow();
      vt = VT_R4|VT_BYREF;
    }
    pfltVal = pfSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ double* pdblSrc) ATLVARIANT_THROW()
  {
    if (vt != (VT_R8|VT_BYREF))
    {
      ClearThrow();
      vt = VT_R8|VT_BYREF;
    }
    pdblVal = pdblSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant& operator=(_In_ const SAFEARRAY* pSrc) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT(pSrc != nullptr);

    if (pSrc == nullptr)
    {
      ClearThrow();
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
    else
    {
      ClearThrow();
      LPSAFEARRAY pCopy{nullptr};
#pragma warning(suppress: 26465 26486 26492)
      HRESULT hr{SafeArrayCopy(const_cast<LPSAFEARRAY>(pSrc), &pCopy)};
      if (SUCCEEDED(hr))
      {
#pragma warning(suppress: 26465 26492)
        hr = ATL::AtlSafeArrayGetActualVartype(const_cast<LPSAFEARRAY>(pSrc), &vt);
        if (SUCCEEDED(hr))
        {
          vt |= VT_ARRAY;
          parray = pCopy;
        }
        else
        {
          vt = VT_ERROR;
          scode = hr;
        }
      }
      else
      {
        vt = VT_ERROR;
        scode = hr;
      }

    #ifndef _ATL_NO_VARIANT_THROW
      if (FAILED(hr))
      {
        if (hr == E_OUTOFMEMORY)
        {
          AtlThrow(E_OUTOFMEMORY);
        }
        else
        {
#pragma warning(suppress: 26477 26496)
          ATLENSURE_THROW(FALSE, hr);
        }
      }
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }

    return *this;
  }

#pragma warning(suppress: 26487)
  bool operator==(_In_ const PROPVARIANT& varSrc) const noexcept
  {
    //PropVariants not equal if types don't match
    if (vt != varSrc.vt)
      return false;
    return PropVariantCompareEx(*this, varSrc, PVCU_DEFAULT, PVCF_DEFAULT) == 0;
  }

#pragma warning(suppress: 26487)
  bool operator!=(_In_ const PROPVARIANT& varSrc) const noexcept
  {
    return !operator==(varSrc);
  }

#pragma warning(suppress: 26487)
  bool operator<(_In_ const PROPVARIANT& varSrc) const noexcept
  {
    if (vt == VT_NULL && varSrc.vt == VT_NULL)
      return false;
    return PropVariantCompareEx(*this, varSrc, PVCU_DEFAULT, PVCF_DEFAULT) == -1;
  }

#pragma warning(suppress: 26487)
  bool operator>(_In_ const PROPVARIANT& varSrc) const noexcept
  {
    if (vt == VT_NULL && varSrc.vt == VT_NULL)
      return false;
    return PropVariantCompareEx(*this, varSrc, PVCU_DEFAULT, PVCF_DEFAULT) == 1;
  }

  HRESULT Clear() noexcept
  {
    return PropVariantClear(this);
  }

  HRESULT Copy(_In_ const PROPVARIANT* pSrc) noexcept
  { 
    return PropVariantCopy(this, pSrc);
  }
 
  //copy PROPVARIANT to BSTR
#pragma warning(suppress: 26487)
  HRESULT CopyTo(_Outptr_result_z_ BSTR* pstrDest) const noexcept
  {
#pragma warning(suppress: 26477)
    ATLASSERT((pstrDest != nullptr) && (vt == VT_BSTR));
    HRESULT hr{E_POINTER};
    if ((pstrDest != nullptr) && (vt == VT_BSTR))
    {
#pragma warning(suppress: 26490)
      *pstrDest = SysAllocStringByteLen(reinterpret_cast<LPCSTR>(bstrVal), SysStringByteLen(bstrVal));
      if (*pstrDest == nullptr)
        hr = E_OUTOFMEMORY;
      else
        hr = S_OK;
    }
    else if (vt != VT_BSTR)
      hr = DISP_E_TYPEMISMATCH;

    return hr;
  }

  HRESULT Attach(_In_ PROPVARIANT* pSrc)
  {
    if (pSrc == nullptr)
      return E_INVALIDARG;

    HRESULT hr{S_OK};
    if (this != pSrc)
    {
      //Clear out the propvariant
      hr = Clear();
      if (SUCCEEDED(hr))
      {
        //Copy the contents and give control to CComPropVariant
        Checked::memcpy_s(this, sizeof(CComPropVariant), pSrc, sizeof(PROPVARIANT));
        pSrc->vt = VT_EMPTY;
        hr = S_OK;
      }
    }
    return hr;
  }

  HRESULT Detach(_Inout_ PROPVARIANT* pDest)
  {
#pragma warning(suppress: 26477)
    ATLASSERT(pDest != nullptr);
    if (pDest == nullptr)
      return E_POINTER;

    //Clear out the variant
    HRESULT hr{PropVariantClear(pDest)};
    if (SUCCEEDED(hr))
    {
      //Copy the contents and remove control from CComPropVariant
      Checked::memcpy_s(pDest, sizeof(PROPVARIANT), this, sizeof(PROPVARIANT));
      vt = VT_EMPTY;
      hr = S_OK;
    }
    return hr;
  }

  HRESULT ChangeType(_In_ VARTYPE vtNew, _In_opt_ const PROPVARIANT* pSrc = nullptr) noexcept
  {
#pragma warning(suppress: 26430)
    const PROPVARIANT* pVar{pSrc};
    //Convert in place if pSrc is nullptr
    if (pVar == nullptr)
      pVar = this;
#pragma warning(suppress: 26486)
    return PropVariantChangeType(this, *pVar, 0, vtNew);
  }

//Implementation
protected:
  void ClearThrow() ATLVARIANT_THROW()
  {
    const HRESULT hr{Clear()};
#pragma warning(suppress: 26477)
    ATLASSERT(SUCCEEDED(hr));
  #ifndef _ATL_NO_VARIANT_THROW
    if (FAILED(hr))
      AtlThrow(hr);
  #else
    UNREFERENCED_PARAMETER(hr);
  #endif //#ifndef _ATL_NO_VARIANT_THROW
  }

  void InternalCopy(_In_ const PROPVARIANT* pSrc) ATLVARIANT_THROW()
  {
    const HRESULT hr{Copy(pSrc)};
    if (FAILED(hr))
    {
      vt = VT_ERROR;
      scode = hr;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(hr);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }
};


//A wrapper class for PROPVARIANT where management of the resource
//is left externally managed
class CComPropVariant2 : public PROPVARIANT
{
public:
//Constructors / Destructors
  CComPropVariant2() noexcept
  {
    //Make sure that propvariant data is initialized to 0
    memset(this, 0, sizeof(PROPVARIANT));
    PropVariantInit(this);
  }

  CComPropVariant2(_In_ const CComPropVariant2& varSrc)
  {
    Checked::memcpy_s(this, sizeof(CComPropVariant2), &varSrc, sizeof(CComPropVariant2));
  }

#pragma warning(suppress: 26439)
  CComPropVariant2(_In_ CComPropVariant2&& varSrc)
  {
    Checked::memcpy_s(this, sizeof(CComPropVariant2), &varSrc, sizeof(CComPropVariant2));
    memset(&varSrc, 0, sizeof(PROPVARIANT));
    PropVariantInit(&varSrc);
  }

  ~CComPropVariant2() noexcept
  {
    const HRESULT hr{Clear()};
#pragma warning(suppress: 26477)
    ATLASSERT(SUCCEEDED(hr));
    UNREFERENCED_PARAMETER(hr);
  }

  CComPropVariant2(_In_ const PROPVARIANT& varSrc)
  {
    Checked::memcpy_s(this, sizeof(PROPVARIANT), &varSrc, sizeof(PROPVARIANT));
  }

  CComPropVariant2(_In_ PROPVARIANT* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_VARIANT|VT_VECTOR;
    capropvar.pElems = pElems;
    capropvar.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ PROPVARIANT* pSrc) noexcept
  {
    vt = VT_VARIANT|VT_BYREF;
    pvarVal = pSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_z_ LPSTR lpszSrc) noexcept
  {
    vt = VT_LPSTR;
    pszVal = lpszSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ LPSTR* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_LPSTR|VT_VECTOR;
    calpstr.pElems = pElems;
    calpstr.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_z_ LPOLESTR lpszSrc, _In_ VARTYPE vtSrc = VT_LPWSTR) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_LPWSTR) || (vtSrc == VT_BSTR));
    if (vtSrc == VT_LPWSTR)
    {
      vt = VT_LPWSTR;
      pwszVal = lpszSrc;
    }
    else if (vtSrc == VT_BSTR)
    {
      vt = vtSrc;
      bstrVal = lpszSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ LPOLESTR* pElems, _In_ ULONG cElems, _In_ VARTYPE vtSrc = VT_LPWSTR|VT_VECTOR) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == (VT_LPWSTR|VT_VECTOR)) || (vtSrc == (VT_BSTR|VT_VECTOR)));
    if (vtSrc == (VT_LPWSTR|VT_VECTOR))
    {
      vt = vtSrc;
      calpwstr.pElems = pElems;
      calpwstr.cElems = cElems;
    }
    else if (vtSrc == (VT_BSTR|VT_VECTOR))
    {
      vt = vtSrc;
      cabstr.pElems = pElems;
      cabstr.cElems = cElems;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ BSTR* lpszSrc) noexcept
  {
    vt = VT_BSTR|VT_BYREF;
    pbstrVal = lpszSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ const FILETIME& ft) noexcept
  {
    vt = VT_FILETIME;
    filetime = ft;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ bool bSrc) noexcept
  {
    vt = VT_BOOL;
    boolVal = bSrc ? VARIANT_TRUE : VARIANT_FALSE;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ int nSrc, _In_ VARTYPE vtSrc = VT_I4) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_I4) || (vtSrc == VT_INT));
    if ((vtSrc == VT_I4) || (vtSrc == VT_INT))
    {
      vt = vtSrc;
      intVal = nSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ BYTE nSrc) noexcept
  {
    vt = VT_UI1;
    bVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ BYTE* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_UI1|VT_VECTOR;
    caub.pElems = pElems;
    caub.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ char cSrc) noexcept
  {
    vt = VT_I1;
    cVal = cSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ char* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_I1|VT_VECTOR;
    cac.pElems = pElems;
    cac.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ short nSrc) noexcept
  {
    vt = VT_I2;
    iVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ short* pnSrc, _In_ VARTYPE vtSrc = VT_I2|VT_BYREF) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == (VT_I2|VT_BYREF)) || (vtSrc == (VT_BOOL|VT_BYREF)));
    if (vtSrc == (VT_I2|VT_BYREF))
    {
      vt = vtSrc;
      piVal = pnSrc;
    }
    else if (vtSrc == (VT_BOOL|VT_BYREF))
    {
      vt = vtSrc;
      pboolVal = pnSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ short* pElems, _In_ ULONG cElems, _In_ VARTYPE vtSrc = VT_I2|VT_VECTOR) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == (VT_I2|VT_VECTOR)) || (vtSrc == (VT_BOOL|VT_VECTOR)));
    if (vtSrc == (VT_I2|VT_VECTOR))
    {
      vt = vtSrc;
      cai.pElems = pElems;
      cai.cElems = cElems;
    }
    else if (vtSrc == (VT_BOOL|VT_VECTOR))
    {
      vt = vtSrc;
      cabool.pElems = pElems;
      cabool.cElems = cElems;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ long nSrc, _In_ VARTYPE vtSrc = VT_I4) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_I4) || (vtSrc == VT_ERROR));
    if (vtSrc == VT_I4)
    {
      vt = vtSrc;
      lVal = nSrc;
    }
    else if (vtSrc == VT_ERROR)
    {
      vt = vtSrc;
      scode = nSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ long* pElems, _In_ ULONG cElems, _In_ VARTYPE vtSrc = VT_I4|VT_VECTOR) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == (VT_I4|VT_VECTOR)) || (vtSrc == (VT_ERROR|VT_VECTOR)));
    if (vtSrc == (VT_I4|VT_VECTOR))
    {
      vt = vtSrc;
      cal.pElems = pElems;
      cal.cElems = cElems;
    }
    else if (vtSrc == (VT_ERROR|VT_VECTOR))
    {
      vt = vtSrc;
      cascode.pElems = pElems;
      cascode.cElems = cElems;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ float fltSrc) noexcept
  {
    vt = VT_R4;
    fltVal = fltSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ float* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_R4|VT_VECTOR;
    caflt.pElems = pElems;
    caflt.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ double dblSrc, _In_ VARTYPE vtSrc = VT_R8) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_R8) || (vtSrc == VT_DATE));
    if (vtSrc == VT_R8)
    {
      vt = vtSrc;
      dblVal = dblSrc;
    }
    else if (vtSrc == VT_DATE)
    {
      vt = vtSrc;
      date = dblSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ double* pElems, _In_ ULONG cElems, _In_ VARTYPE vtSrc = VT_R8|VT_VECTOR) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == (VT_R8|VT_VECTOR)) || (vtSrc == (VT_DATE|VT_VECTOR)));
    if (vtSrc == (VT_R8|VT_VECTOR))
    {
      vt = vtSrc;
      cadbl.pElems = pElems;
      cadbl.cElems = cElems;
    }
    else if (vtSrc == (VT_DATE|VT_VECTOR))
    {
      vt = vtSrc;
      cadate.pElems = pElems;
      cadate.cElems = cElems;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ DECIMAL& decSrc)
  {
    Checked::memcpy_s(&decVal, sizeof(decVal), &decSrc, sizeof(decSrc));
    vt = VT_DECIMAL; //Note it is important we set the vt value after we do the memcpy
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ BLOB& blobSrc) noexcept
  {
    vt = VT_BLOB;
    blob.cbSize = blobSrc.cbSize;
    blob.pBlobData = blobSrc.pBlobData;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ BSTRBLOB& bstrblobSrc) noexcept
  {
    vt = VT_BSTR_BLOB;
    bstrblobVal.cbSize = bstrblobSrc.cbSize;
    bstrblobVal.pData = bstrblobSrc.pData;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ BSTRBLOB* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_BSTR_BLOB|VT_VECTOR;
    cabstrblob.pElems = pElems;
    cabstrblob.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ LARGE_INTEGER nSrc) noexcept
  {
    vt = VT_I8;
    hVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ LARGE_INTEGER* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_I8|VT_VECTOR;
    cah.pElems = pElems;
    cah.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ ULARGE_INTEGER nSrc) noexcept
  {
    vt = VT_UI8;
    uhVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ ULARGE_INTEGER* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_UI8|VT_VECTOR;
    cauh.pElems = pElems;
    cauh.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ CY cySrc) noexcept
  {
    vt = VT_CY;
    cyVal.Hi = cySrc.Hi;
    cyVal.Lo = cySrc.Lo;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ CY* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_CY|VT_VECTOR;
    cacy.pElems = pElems;
    cacy.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ IDispatch* pSrc) noexcept
  {
    vt = VT_DISPATCH;
    pdispVal = pSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ IDispatch** ppSrc) noexcept
  {
    vt = VT_DISPATCH|VT_BYREF;
    ppdispVal = ppSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ IUnknown* pSrc) noexcept
  {
    vt = VT_UNKNOWN;
    punkVal = pSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ IUnknown** ppSrc) noexcept
  {
    vt = VT_UNKNOWN|VT_BYREF;
    ppunkVal = ppSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ IStream* pSrc, _In_ VARTYPE vtSrc = VT_STREAM) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_STREAM) || (vtSrc == VT_STREAMED_OBJECT));
    if ((vtSrc == VT_STREAM) || (vtSrc == VT_STREAMED_OBJECT))
    {
      vt = vtSrc;
      pStream = pSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ IStorage* pSrc, _In_ VARTYPE vtSrc = VT_STORAGE) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_STORAGE) || (vtSrc == VT_STORED_OBJECT));
    if ((vtSrc == VT_STORAGE) || (vtSrc == VT_STORED_OBJECT))
    {
      vt = vtSrc;
      pStorage = pSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ VERSIONEDSTREAM* pSrc) noexcept
  {
    vt = VT_VERSIONED_STREAM;
    pVersionedStream = pSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ unsigned short nSrc) noexcept
  {
    vt = VT_UI2;
    uiVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ unsigned short* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_UI2|VT_VECTOR;
    caui.pElems = pElems;
    caui.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ unsigned long nSrc) noexcept
  {
    vt = VT_UI4;
    ulVal = nSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ unsigned long* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_UI4|VT_VECTOR;
    caul.pElems = pElems;
    caul.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ unsigned int nSrc, _In_ VARTYPE vtSrc = VT_UI4) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == VT_UI4) || (vtSrc == VT_UINT));
    if ((vtSrc == VT_UI4) || (vtSrc == VT_UINT))
    {
      vt = vtSrc;
      uintVal = nSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ SAFEARRAY* pSrc) ATLVARIANT_THROW()
  {
    vt = VT_EMPTY;
    *this = pSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ SAFEARRAY** ppSrc) noexcept
  {
    vt = VT_ARRAY|VT_BYREF;
    pparray = ppSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ CLSID* pSrc) noexcept
  {
    vt = VT_CLSID;
    puuid = pSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ CLSID* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_CLSID|VT_VECTOR;
    cauuid.pElems = pElems;
    cauuid.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ CLIPDATA* pSrc) noexcept
  {
    vt = VT_CF;
    pclipdata = pSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ CLIPDATA* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_CF|VT_VECTOR;
    caclipdata.pElems = pElems;
    caclipdata.cElems = cElems;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ DECIMAL* pSrc) noexcept
  {
    vt = VT_DECIMAL|VT_BYREF;
    pdecVal = pSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ BYTE* pbSrc) noexcept
  {
    vt = VT_UI1|VT_BYREF;
    pbVal = pbSrc;
  }

#ifdef _NATIVE_WCHAR_T_DEFINED
#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ USHORT* pnSrc) noexcept
  {
    vt = VT_UI2|VT_BYREF;
    puiVal = pnSrc;
  }
#endif //#ifdef _NATIVE_WCHAR_T_DEFINED

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ int* pnSrc) noexcept
  {
    vt = VT_I4|VT_BYREF;
    pintVal = pnSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ UINT* pnSrc) noexcept
  {
    vt = VT_UI4|VT_BYREF;
    puintVal = pnSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ long* pnSrc, _In_ VARTYPE vtSrc = VT_I4|VT_BYREF) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT((vtSrc == (VT_I4|VT_BYREF)) || (vtSrc == (VT_ERROR|VT_BYREF)));
    if (vtSrc == (VT_I4|VT_BYREF))
    {
      vt = vtSrc;
      plVal = pnSrc;
    }
    else if (vtSrc == (VT_ERROR|VT_BYREF))
    {
      vt = vtSrc;
      pscode = pnSrc;
    }
    else
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ ULONG* pnSrc) noexcept
  {
    vt = VT_UI4|VT_BYREF;
    pulVal = pnSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ CY* pcySrc) noexcept
  {
    vt = VT_CY|VT_BYREF;
    pcyVal = pcySrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ float* pfSrc) noexcept
  {
    vt = VT_R4|VT_BYREF;
    pfltVal = pfSrc;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_opt_ double* pdblSrc) noexcept
  {
    vt = VT_R8|VT_BYREF;
    pdblVal = pdblSrc;
  }

#pragma warning(suppress: 26434 26487)
  CComPropVariant2& operator=(_In_ const PROPVARIANT& varSrc)
  {
    if (this != &varSrc)
      Checked::memcpy_s(this, sizeof(PROPVARIANT), &varSrc, sizeof(PROPVARIANT));
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ PROPVARIANT* pSrc) noexcept
  {
    vt = VT_VARIANT|VT_BYREF;
    pvarVal = pSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_z_ LPSTR lpszSrc) noexcept
  {
    vt = VT_LPSTR;
    pszVal = lpszSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ bool bSrc) noexcept
  {
    vt = VT_BOOL;
    boolVal = bSrc ? VARIANT_TRUE : VARIANT_FALSE;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ const FILETIME& nSrc) noexcept
  {
    vt = VT_FILETIME;
    filetime = nSrc;
    return *this;
  }

#pragma warning(suppress: 26495)
  CComPropVariant2(_In_ FILETIME* pElems, _In_ ULONG cElems) noexcept
  {
    vt = VT_FILETIME|VT_VECTOR;
    cafiletime.pElems = pElems;
    cafiletime.cElems = cElems;
  }

#pragma warning(suppress: 26456)
  CComPropVariant2& operator=(_In_ const CComPropVariant2& varSrc)
  {
    if (this != &varSrc)
      Checked::memcpy_s(this, sizeof(CComPropVariant2), &varSrc, sizeof(CComPropVariant2));
    return *this;
  }

#pragma warning(suppress: 26439 26456)
  CComPropVariant2& operator=(_In_ CComPropVariant2&& varSrc)
  {
    Checked::memcpy_s(this, sizeof(CComPropVariant2), &varSrc, sizeof(CComPropVariant2));
    memset(&varSrc, 0, sizeof(PROPVARIANT));
    PropVariantInit(&varSrc);
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ int nSrc) noexcept
  {
    vt = VT_I4;
    intVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ BYTE nSrc) noexcept
  {
    vt = VT_UI1;
    bVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ char cSrc) noexcept
  {
    vt = VT_I1;
    cVal = cSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ short nSrc) noexcept
  {
    vt = VT_I2;
    iVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ long nSrc) noexcept
  {
    vt = VT_I4;
    lVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ float fltSrc) noexcept
  {
    vt = VT_R4;
    fltVal = fltSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ double dblSrc) noexcept
  {
    vt = VT_R8;
    dblVal = dblSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ DECIMAL& decSrc)
  {
    Checked::memcpy_s(&decVal, sizeof(decVal), &decSrc, sizeof(decSrc));
    vt = VT_DECIMAL; //Note it is important we set the vt value after we do the memcpy
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ BLOB& blobSrc) noexcept
  {
    vt = VT_BLOB;
    blob.cbSize = blobSrc.cbSize;
    blob.pBlobData = blobSrc.pBlobData;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ BSTRBLOB& bstrblobSrc) noexcept
  {
    vt = VT_BSTR_BLOB;
    bstrblobVal.cbSize = bstrblobSrc.cbSize;
    bstrblobVal.pData = bstrblobSrc.pData;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ CY cySrc) noexcept
  {
    vt = VT_CY;
    cyVal.Hi = cySrc.Hi;
    cyVal.Lo = cySrc.Lo;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ unsigned short nSrc) noexcept
  {
    vt = VT_UI2;
    uiVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ unsigned long nSrc) noexcept
  {
    vt = VT_UI4;
    ulVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ unsigned int nSrc) noexcept
  {
    vt = VT_UI4;
    uintVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ LARGE_INTEGER nSrc) noexcept
  {
    vt = VT_I8;
    hVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ ULARGE_INTEGER nSrc) noexcept
  {
    vt = VT_UI8;
    uhVal = nSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ CLSID* pSrc) noexcept
  {
    vt = VT_CLSID;
    puuid = pSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ CLIPDATA* pSrc) noexcept
  {
    vt = VT_CF;
    pclipdata = pSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ DECIMAL* pSrc) noexcept
  {
    vt = VT_DECIMAL|VT_BYREF;
    pdecVal = pSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ IDispatch* pSrc) noexcept
  {
    vt = VT_DISPATCH;
    pdispVal = pSrc;
    return *this;
  }

#pragma warning(suppress: 26434 26487)
  CComPropVariant2& operator=(_In_opt_ IDispatch** ppSrc) noexcept
  {
    vt = VT_DISPATCH|VT_BYREF;
    ppdispVal = ppSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ IUnknown* pSrc) noexcept
  {
    vt = VT_UNKNOWN;
    punkVal = pSrc;
    return *this;
  }

#pragma warning(suppress: 26434 26487)
  CComPropVariant2& operator=(_In_opt_ IUnknown** ppSrc) noexcept
  {
    vt = VT_UNKNOWN|VT_BYREF;
    ppunkVal = ppSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ LPVERSIONEDSTREAM pSrc) noexcept
  {
    vt = VT_VERSIONED_STREAM;
    pVersionedStream = pSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ BYTE* pbSrc) noexcept
  {
    vt = VT_UI1|VT_BYREF;
    pbVal = pbSrc;
    return *this;
  }

#ifdef _NATIVE_WCHAR_T_DEFINED
#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ USHORT* pnSrc) noexcept
  {
    vt = VT_UI2|VT_BYREF;
    puiVal = pnSrc;
    return *this;
  }
#endif //#ifdef _NATIVE_WCHAR_T_DEFINED

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ int* pnSrc) noexcept
  {
    vt = VT_I4|VT_BYREF;
    pintVal = pnSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ UINT* pnSrc) noexcept
  {
    vt = VT_UI4|VT_BYREF;
    puintVal = pnSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ long* pnSrc) noexcept
  {
    vt = VT_I4|VT_BYREF;
    plVal = pnSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ ULONG* pnSrc) noexcept
  {
    vt = VT_UI4|VT_BYREF;
    pulVal = pnSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ CY* pVal) noexcept
  {
    vt = VT_CY|VT_BYREF;
    pcyVal = pVal;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ float* pfSrc) noexcept
  {
    vt = VT_R4|VT_BYREF;
    pfltVal = pfSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_opt_ double* pdblSrc) noexcept
  {
    vt = VT_R8|VT_BYREF;
    pdblVal = pdblSrc;
    return *this;
  }

#pragma warning(suppress: 26434)
  CComPropVariant2& operator=(_In_ SAFEARRAY* pSrc) ATLVARIANT_THROW()
  {
#pragma warning(suppress: 26477)
    ATLASSERT(pSrc != nullptr);

    if (pSrc == nullptr)
    {
      vt = VT_ERROR;
      scode = E_INVALIDARG;
    #ifndef _ATL_NO_VARIANT_THROW
      AtlThrow(E_INVALIDARG);
    #endif //#ifndef _ATL_NO_VARIANT_THROW
    }
    else
    {
      const HRESULT hr{ATL::AtlSafeArrayGetActualVartype(pSrc, &vt)};
      if (SUCCEEDED(hr))
      {
        vt |= VT_ARRAY;
        parray = pSrc;
      }
      else
      {
        vt = VT_ERROR;
        scode = hr;

      #ifndef _ATL_NO_VARIANT_THROW
        if (FAILED(hr))
        {
#pragma warning(suppress: 26496 26477)
          ATLENSURE_THROW(FALSE, hr);
        }
      #endif //#ifndef _ATL_NO_VARIANT_THROW
      }
    }

    return *this;
  }

#pragma warning(suppress: 26434 26487)
  CComPropVariant2& operator=(_In_ SAFEARRAY** ppSrc) noexcept
  {
    vt = VT_ARRAY|VT_BYREF;
    pparray = ppSrc;
    return *this;
  }

#pragma warning(suppress: 26487)
  bool operator==(_In_ const PROPVARIANT& varSrc) const noexcept
  {
    //PropVariants not equal if types don't match
    if (vt != varSrc.vt)
      return false;
    return PropVariantCompareEx(*this, varSrc, PVCU_DEFAULT, PVCF_DEFAULT) == 0;
  }

#pragma warning(suppress: 26487)
  bool operator!=(_In_ const PROPVARIANT& varSrc) const noexcept
  {
    return !operator==(varSrc);
  }

#pragma warning(suppress: 26487)
  bool operator<(_In_ const PROPVARIANT& varSrc) const noexcept
  {
    if ((vt == VT_NULL) && (varSrc.vt == VT_NULL))
      return false;
    return PropVariantCompareEx(*this, varSrc, PVCU_DEFAULT, PVCF_DEFAULT) == -1;
  }

#pragma warning(suppress: 26487)
  bool operator>(_In_ const PROPVARIANT& varSrc) const noexcept
  {
    if ((vt == VT_NULL) && (varSrc.vt == VT_NULL))
      return false;
    return PropVariantCompareEx(*this, varSrc, PVCU_DEFAULT, PVCF_DEFAULT) == 1;
  }

  HRESULT Clear() noexcept
  {
    PropVariantInit(this);
    return S_OK;
  }

  //copy PROPVARIANT to BSTR
#pragma warning(suppress: 26487)
  HRESULT CopyTo(_Outptr_result_z_ BSTR* pstrDest) const noexcept
  {
#pragma warning(suppress: 26477)
    ATLASSERT((pstrDest != nullptr) && (vt == VT_BSTR));
    HRESULT hr{E_POINTER};
    if ((pstrDest != nullptr) && (vt == VT_BSTR))
    {
#pragma warning(suppress: 26490)
      *pstrDest = SysAllocStringByteLen(reinterpret_cast<LPCSTR>(bstrVal), SysStringByteLen(bstrVal));
      if (*pstrDest == nullptr)
        hr = E_OUTOFMEMORY;
      else
        hr = S_OK;
    }
    else if (vt != VT_BSTR)
      hr = DISP_E_TYPEMISMATCH;

    return hr;
  }

  HRESULT Attach(_In_ PROPVARIANT* pSrc)
  {
    if (pSrc == nullptr)
      return E_INVALIDARG;

    if (this != pSrc)
    {
      //Copy the contents and give control to CComPropVariant2
      Checked::memcpy_s(this, sizeof(CComPropVariant2), pSrc, sizeof(PROPVARIANT));
      pSrc->vt = VT_EMPTY;
    }

    return S_OK;
  }

  HRESULT Detach(_Inout_ PROPVARIANT* pDest)
  {
#pragma warning(suppress: 26477)
    ATLASSERT(pDest != nullptr);
    if (pDest == nullptr)
      return E_POINTER;

    //Copy the contents and remove control from CComPropVariant2
    Checked::memcpy_s(pDest, sizeof(PROPVARIANT), this, sizeof(PROPVARIANT));
    vt = VT_EMPTY;

    return S_OK;
  }

  HRESULT ChangeType(_In_ VARTYPE vtNew, _In_opt_ const PROPVARIANT* pSrc = nullptr) noexcept
  {
#pragma warning(suppress: 26430)
    const PROPVARIANT* pVar = pSrc;
    //Convert in place if pSrc is nullptr
    if (pVar == nullptr)
      pVar = this;
#pragma warning(suppress: 26486)
    return PropVariantChangeType(this, *pVar, 0, vtNew);
  }
};


}; //namespace ATL

#endif //#ifndef __PROPVARIANTWRAPPER_H__
